"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateSortFieldsForDocument = exports.getAverageSortId = exports.documentsBinarySearch = void 0;
const Base2N = require("@navpreetdevpuri/base-2-n");
function documentsBinarySearch(model, fieldName, value, sortFieldName, ignoreCases) {
    return __awaiter(this, void 0, void 0, function* () {
        const n = yield model
            .findOne({ [sortFieldName]: { $ne: null } })
            .count()
            .exec();
        let start = 0;
        let end = n - 1;
        let startDoc = yield model
            .findOne({ [sortFieldName]: { $ne: null } }, { [fieldName]: 1, [sortFieldName]: 1 })
            .sort({ [sortFieldName]: 1 })
            .skip(start)
            .exec();
        let midDoc;
        let endDoc = yield model
            .findOne({ [sortFieldName]: { $ne: null } }, { [fieldName]: 1 })
            .sort({ [sortFieldName]: 1 })
            .skip(end)
            .exec();
        let startValue;
        let midValue;
        let endValue;
        while (start <= end) {
            const mid = Math.ceil((start + end) / 2);
            midDoc = yield model
                .findOne({ [sortFieldName]: { $ne: null } }, { [fieldName]: 1 })
                .sort({ [sortFieldName]: 1 })
                .skip(mid)
                .exec();
            startValue = String(startDoc[fieldName]);
            midValue = String(midDoc[fieldName]);
            endValue = String(endDoc[fieldName]);
            startValue = ignoreCases ? startValue.toLowerCase() : startValue;
            midValue = ignoreCases ? midValue.toLowerCase() : midValue;
            endValue = ignoreCases ? endValue.toLowerCase() : endValue;
            if (value < midValue) {
                end = mid - 1;
                endDoc = yield model
                    .findOne({ [sortFieldName]: { $ne: null } }, { [fieldName]: 1 })
                    .sort({ [sortFieldName]: 1 })
                    .skip(end)
                    .exec();
            }
            else if (midValue <= value) {
                start = mid + 1;
                startDoc = yield model
                    .findOne({ [sortFieldName]: { $ne: null } }, { [fieldName]: 1, [sortFieldName]: 1 })
                    .sort({ [sortFieldName]: 1 })
                    .skip(start)
                    .exec();
            }
        }
        if (start === 0) {
            return {
                predecessorSortId: null,
                successorSortId: startDoc[sortFieldName],
            };
        }
        if (start === n) {
            startDoc = yield model
                .findOne({ [sortFieldName]: { $ne: null } }, { [sortFieldName]: 1 })
                .sort({ [sortFieldName]: -1 })
                .exec();
            return {
                predecessorSortId: startDoc[sortFieldName],
                successorSortId: null,
            };
        }
        let predecessorDoc = endDoc;
        const successorDoc = startDoc;
        if (end === start - 1) {
            predecessorDoc = yield model
                .findOne({ [sortFieldName]: { $ne: null } }, { [sortFieldName]: 1 })
                .sort({ [sortFieldName]: 1 })
                .skip(end)
                .exec();
        }
        return {
            predecessorSortId: predecessorDoc[sortFieldName],
            successorSortId: successorDoc[sortFieldName],
        };
    });
}
exports.documentsBinarySearch = documentsBinarySearch;
function getAverageSortId(predecessorSortId, successorSortId, noOfCharsToIncreaseOnSaturation) {
    if (!predecessorSortId) {
        predecessorSortId = "".padEnd(successorSortId.length, "\0");
    }
    if (!successorSortId) {
        successorSortId = "".padEnd(predecessorSortId.length, "\uffff");
    }
    let predecessorNumber;
    let successorNumber;
    if (predecessorSortId.length == predecessorSortId.length) {
        predecessorNumber = new Base2N(predecessorSortId);
        successorNumber = new Base2N(successorSortId);
        const averageNumber = predecessorNumber.average(successorNumber);
        if (averageNumber.toString() != predecessorNumber.toString()) {
            return averageNumber.toString();
        }
        predecessorNumber = new Base2N(predecessorSortId.padEnd(averageNumber.length + noOfCharsToIncreaseOnSaturation, "\0"));
        successorNumber = new Base2N(successorSortId.padEnd(averageNumber.length + noOfCharsToIncreaseOnSaturation, "\0"));
        return predecessorNumber.average(successorNumber).toString();
    }
    const bigger = predecessorSortId.length > successorSortId.length ? predecessorSortId : successorSortId;
    const smaller = successorSortId.length > predecessorSortId.length ? predecessorSortId : successorSortId;
    const biggerNumber = new Base2N(bigger);
    const smallerNumber = new Base2N(smaller.padEnd(bigger.length, "\0"));
    return biggerNumber.average(smallerNumber).toString();
}
exports.getAverageSortId = getAverageSortId;
function updateSortFieldsForDocument({ objectId, model, fieldName, fieldValue, sortFieldName, ignoreCases, noOfCharsToIncreaseOnSaturation, }) {
    return __awaiter(this, void 0, void 0, function* () {
        console.time(`Total time took to update orderId: ${objectId}`);
        const { predecessorSortId, successorSortId } = yield documentsBinarySearch(model, fieldName, fieldValue, sortFieldName, ignoreCases);
        const newSortId = getAverageSortId(predecessorSortId, successorSortId, noOfCharsToIncreaseOnSaturation);
        yield model.updateOne({ _id: objectId }, { $set: { [sortFieldName]: newSortId.toString() } });
        const documentsCountWithSameSortId = yield model
            .find({ [sortFieldName]: newSortId.toString() })
            .count()
            .exec();
        if (documentsCountWithSameSortId > 1) {
            console.log(`mongoose-sort-encrypted-field -> Got collions, retrying... ${objectId}`);
            // Retrigering sortId generation due to collion
            const documentsWithSameSortId = yield model.find({ [sortFieldName]: newSortId.toString() }, { [fieldName]: 1 }).exec();
            for (const document of documentsWithSameSortId) {
                yield model.updateOne({ _id: objectId }, { $set: { [fieldName]: document[fieldName] } });
            }
        }
        console.timeEnd(`Total time took to update orderId: ${objectId}`);
    });
}
exports.updateSortFieldsForDocument = updateSortFieldsForDocument;
